import unittest

from catbuffer_parser.CatsParseException import CatsParseException
from catbuffer_parser.EnumParser import EnumParserFactory, EnumValueParserFactory

from .constants import INVALID_PROPERTY_NAMES, INVALID_USER_TYPE_NAMES, PRIMITIVE_TYPE_TUPLES, VALID_PROPERTY_NAMES, VALID_USER_TYPE_NAMES
from .ParserTestUtils import MultiLineParserTestUtils, ParserFactoryTestUtils, SingleLineParserTestUtils


def primitive_enum_descriptor(size, is_signed):
    return {'type': 'enum', 'signedness': 'signed' if is_signed else 'unsigned', 'size': size}


class EnumParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(EnumParserFactory, self).assert_positives([
            'enum F : uint8', 'enum Foo : uint7', 'enum FooZA09za : uint8', 'enum 8oo : uint9', 'enum $^^$ : uint8'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(EnumParserFactory, self).assert_negatives([
            ' enum F : uint8', 'enum F : uint8 ', 'enum F', 'enum F :', 'enum F : uintA', 'enum F : binary_fixed(8)', 'enum F : Foo'
        ])


class EnumParserTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Assert:
        MultiLineParserTestUtils(EnumParserFactory, self).assert_parse(line, expected_result)

    def _assert_parse_exception(self, line):
        # Assert:
        MultiLineParserTestUtils(EnumParserFactory, self).assert_parse_exception(line)

    def test_parser_exposes_custom_factories(self):
        # Act
        parser = EnumParserFactory().create()

        # Assert
        self.assertEqual(1, len(parser.factories()))

    def test_can_parse_type_declaration(self):
        for primitive_tuple in PRIMITIVE_TYPE_TUPLES:
            # Act + Assert:
            self._assert_parse(
                'enum Colors : {0}'.format(primitive_tuple[0]),
                ('Colors', {'type': 'enum', 'size': primitive_tuple[1], 'signedness': primitive_tuple[2], 'values': []}))

    def test_cannot_parse_enum_declaration_with_invalid_base(self):
        for base_type in ['uint7', 'uint9']:
            # Act + Assert:
            self._assert_parse_exception('enum Colors : {0}'.format(base_type))

    def test_enum_names_must_have_type_name_semantics(self):
        # Assert:
        MultiLineParserTestUtils(EnumParserFactory, self).assert_naming(
            'enum {0} : uint8',
            VALID_USER_TYPE_NAMES,
            INVALID_USER_TYPE_NAMES)

    def test_can_append_value(self):
        # Arrange:
        parser = EnumParserFactory().create()

        # Act:
        parser.process_line('enum Colors : uint16')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Colors', {**primitive_enum_descriptor(2, False), 'values': [{'name': 'foo'}, {'name': 'bar'}]}), result)

    def test_cannot_append_multiple_properties_with_same_name(self):
        # Arrange:
        parser = EnumParserFactory().create()

        # Act:
        parser.process_line('enum Colors : uint16')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar'})

        # Assert:
        with self.assertRaises(CatsParseException):
            parser.append({'name': 'foo'})

    def test_can_append_multiple_properties_with_same_value(self):
        # Arrange:
        parser = EnumParserFactory().create()

        # Act:
        parser.process_line('enum Colors : int16')
        parser.append({'name': 'foo', 'value': 2})
        parser.append({'name': 'bar', 'value': 2})
        result = parser.commit()

        # Assert:
        self.assertEqual(
            ('Colors', {**primitive_enum_descriptor(2, True), 'values': [{'name': 'foo', 'value': 2}, {'name': 'bar', 'value': 2}]}),
            result)


class EnumValueParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(EnumValueParserFactory, self).assert_positives([
            'foo = bar', 'foo = BAR', 'fzaZa09 = d', 'f = ccc', 'foo = fazFZA90', '$$$ = ^^^'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(EnumValueParserFactory, self).assert_negatives([
            ' foo = bar', 'foo = bar ', 'foo = ', '= bar', 'foo = array(bar, baz)'
        ])


class EnumValueParserTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Assert:
        SingleLineParserTestUtils(EnumValueParserFactory, self).assert_parse(line, expected_result)

    def test_can_parse_dec_declaration(self):
        # Act + Assert:
        self._assert_parse(
            'red = 12',
            {'name': 'red', 'value': 12})

    def test_can_parse_hex_declaration(self):
        # Act + Assert:
        self._assert_parse(
            'red = 0x11',
            {'name': 'red', 'value': 17})

    def test_cannot_parse_non_numeric_declaration(self):
        # Act + Assert:
        SingleLineParserTestUtils(EnumValueParserFactory, self).assert_parse_exception('red = uint16', ValueError)

    def test_member_names_must_have_property_name_semantics(self):
        # Assert:
        SingleLineParserTestUtils(EnumValueParserFactory, self).assert_naming('{0} = 12', VALID_PROPERTY_NAMES, INVALID_PROPERTY_NAMES)



from .CatsParseException import CatsParseException
from .CompositeTypeParser import CompositeTypeParser
from .parserutils import (is_builtin, is_dec_or_hex, is_primitive, parse_builtin, parse_dec_or_hex, require_property_name,
                          require_user_type_name)
from .RegexParserFactory import RegexParserFactory

# region StructParser(Factory)


class StructParser(CompositeTypeParser):
    """Parser for `struct` statements"""
    def __init__(self, regex):
        super().__init__(regex, [
            StructConstParserFactory(),
            StructInlineParserFactory(),
            StructScalarMemberParserFactory(),
            StructArrayMemberParserFactory()
        ])

    def process_line(self, line):
        match = self.regex.match(line)
        self.type_name = require_user_type_name(match.group(1))
        self.type_descriptor = {'type': 'struct', 'layout': []}

    def append(self, property_type_descriptor):
        self._require_no_array_with_fill_disposition()

        if 'size' in property_type_descriptor:
            self._require_known_property(property_type_descriptor['size'])

        descriptor_uid = self._get_descriptor_uid(property_type_descriptor)
        if descriptor_uid[0]:
            self._require_unknown_property(descriptor_uid)

        self.type_descriptor['layout'].append(property_type_descriptor)

    def commit(self):
        for property_type_descriptor in self.type_descriptor['layout']:
            if 'condition' in property_type_descriptor:
                self._require_known_property(property_type_descriptor['condition'], False)

        return super().commit()

    def _require_no_array_with_fill_disposition(self):
        layout = self.type_descriptor['layout']
        if not layout:
            return

        last_property = layout[-1]
        if 'fill' == last_property.get('disposition'):
            raise CatsParseException('array property with fill disposition "{0}" must be last property'.format(last_property['name']))

    def _require_known_property(self, property_name, allow_numeric=True):
        # size can be a constant represented by a numeric type
        if allow_numeric and not isinstance(property_name, str):
            return

        if all('name' not in property_type_descriptor or property_name != property_type_descriptor['name']
               for property_type_descriptor in self.type_descriptor['layout']):
            raise CatsParseException('no definition for referenced property "{0}"'.format(property_name))

    def _require_unknown_property(self, descriptor_uid):
        if any(descriptor_uid == self._get_descriptor_uid(property_type_descriptor)
               for property_type_descriptor in self.type_descriptor['layout']):
            raise CatsParseException('duplicate definition for property "{0}"'.format(descriptor_uid))

    @staticmethod
    def _get_descriptor_uid(descriptor):
        return (descriptor.get('name'), descriptor.get('disposition'))


class StructParserFactory(RegexParserFactory):
    """Factory for creating struct parsers"""
    def __init__(self):
        super().__init__(r'struct (\S+)', StructParser)

# endregion

# region StructConstParser(Factory)


class StructConstParser:
    """Parser for const struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        type_name = match.group(1)

        const_descriptor = {
            'name': require_property_name(match.group(2)),
            'disposition': 'const',
            'value': parse_dec_or_hex(match.group(3))
        }

        if is_primitive(type_name):
            const_descriptor = {**const_descriptor, **parse_builtin(type_name)}
        else:
            const_descriptor['type'] = require_user_type_name(type_name)

        return const_descriptor


class StructConstParserFactory(RegexParserFactory):
    """Factory for creating struct const parsers"""
    def __init__(self):
        super().__init__(r'const (\S+) (\S+) = (\S+)', StructConstParser)


# endregion

# region StructInlineParser(Factory)

class StructInlineParser:
    """Parser for inline struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        # type is resolved to exist upstream, so its naming doesn't need to be checked here
        match = self.regex.match(line)
        return {'type': match.group(1), 'disposition': 'inline'}


class StructInlineParserFactory(RegexParserFactory):
    """Factory for creating struct inline parsers"""
    def __init__(self):
        super().__init__(r'inline (\S+)', StructInlineParser)

# endregion

# region StructScalarMemberParser(Factory)


class StructScalarMemberParser:
    """Parser for non-inline scalar struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        linked_type_name = match.group(2)

        # type is resolved to exist upstream, so its naming doesn't need to be checked here
        if is_builtin(linked_type_name):
            property_type_descriptor = parse_builtin(linked_type_name)  # reduce builtins to byte
        else:
            property_type_descriptor = {'type': linked_type_name}

        if match.group(3):
            property_type_descriptor['condition'] = match.group(4)
            property_type_descriptor['condition_operation'] = match.group(5)
            property_type_descriptor['condition_value'] = match.group(6)

        property_type_descriptor['name'] = require_property_name(match.group(1))
        return property_type_descriptor


class StructScalarMemberParserFactory(RegexParserFactory):
    """Factory for creating struct scalar member parsers"""
    def __init__(self):
        super().__init__(r'(\S+) = (\S+)( if (\S+) (equals|has) (\S+))?', StructScalarMemberParser)

# endregion

# region StructArrayMemberParser(Factory)


class StructArrayMemberParser:
    """Parser for non-inline array struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)

        # type is resolved to exist upstream, so its naming doesn't need to be checked here
        property_type_descriptor = {'type': match.group(2)}

        # size can be interpreted in different ways for count-based arrays
        # size must be a field reference for size-based arrays
        array_size = match.group(4)
        if not match.group(3):
            if is_dec_or_hex(array_size):
                array_size = parse_dec_or_hex(array_size)

            if '__FILL__' == array_size:
                array_size = 0
                property_type_descriptor['disposition'] = 'fill'
        else:
            property_type_descriptor['disposition'] = 'var'

        property_type_descriptor['size'] = array_size

        if match.group(5):
            property_type_descriptor['sort_key'] = match.group(6)

        property_type_descriptor['name'] = require_property_name(match.group(1))
        return property_type_descriptor


class StructArrayMemberParserFactory(RegexParserFactory):
    """Factory for creating struct array member parsers"""
    def __init__(self):
        super().__init__(r'(\S+) = array\((\S+), (size=)?(\S+)(, sort_key=(\S+))?\)', StructArrayMemberParser)

# endregion



import unittest

from catbuffer_parser.CatsParseException import CatsParseException
from catbuffer_parser.parserutils import (is_builtin, is_dec_or_hex, is_primitive, parse_builtin, parse_dec_or_hex, require_primitive,
                                          require_property_name, require_user_type_name)

from .constants import (BUILTIN_TYPE_TUPLES, INT_TYPE_TUPLES, INVALID_PRIMITIVE_NAMES, INVALID_PROPERTY_NAMES, INVALID_USER_TYPE_NAMES,
                        UINT_TYPE_TUPLES, VALID_PRIMITIVE_NAMES, VALID_PROPERTY_NAMES, VALID_USER_TYPE_NAMES)

# region naming conventions


class RequireUserTypeNameTest(unittest.TestCase):
    def test_nothrow_for_positives(self):
        for string in VALID_USER_TYPE_NAMES:
            # Act:
            result = require_user_type_name(string)

            # Assert:
            self.assertEqual(string, result)

    def test_throw_for_negatives(self):
        for string in INVALID_USER_TYPE_NAMES:
            # Act:
            with self.assertRaises(CatsParseException):
                require_user_type_name(string)


class RequirePropertyNameTest(unittest.TestCase):
    def test_nothrow_for_positives(self):
        for string in VALID_PROPERTY_NAMES:
            # Act:
            result = require_property_name(string)

            # Assert:
            self.assertEqual(string, result)

    def test_throw_for_negatives(self):
        for string in INVALID_PROPERTY_NAMES:
            # Act:
            with self.assertRaises(CatsParseException):
                require_property_name(string)

# endregion

# region primitive


class IsPrimitiveTest(unittest.TestCase):
    def test_true_for_positives(self):
        for string in VALID_PRIMITIVE_NAMES:
            # Act:
            result = is_primitive(string)

            # Assert:
            self.assertTrue(result)

    def test_false_for_negatives(self):
        for string in INVALID_PRIMITIVE_NAMES:
            # Act:
            result = is_primitive(string)

            # Assert:
            self.assertFalse(result)


class RequirePrimitiveTest(unittest.TestCase):
    def test_nothrow_for_positives(self):
        for string in VALID_PRIMITIVE_NAMES:
            # Act:
            result = require_primitive(string)

            # Assert:
            self.assertEqual(string, result)

    def test_throw_for_negatives(self):
        for string in INVALID_PRIMITIVE_NAMES:
            # Act:
            with self.assertRaises(CatsParseException):
                require_primitive(string)

# endregion

# region dec or hex


INVALID_NUMERIC_STRINGS = ['AFE', '0x8Y8', 'p', '&']


class IsDecOrHexTest(unittest.TestCase):
    def test_true_for_positives(self):
        for string in ['10', '123', '0x10', '0x123', '0xAFE']:
            # Act:
            result = is_dec_or_hex(string)

            # Assert:
            self.assertTrue(result)

    def test_false_for_negatives(self):
        for string in INVALID_NUMERIC_STRINGS:
            # Act:
            result = is_dec_or_hex(string)

            # Assert:
            self.assertFalse(result)


class ParseDecOrHexTest(unittest.TestCase):
    def test_can_parse_dec(self):
        # Act + Assert:
        self.assertEqual(10, parse_dec_or_hex('10'))
        self.assertEqual(123, parse_dec_or_hex('123'))

    def test_can_parse_hex(self):
        # Act + Assert:
        self.assertEqual(0x10, parse_dec_or_hex('0x10'))
        self.assertEqual(0x123, parse_dec_or_hex('0x123'))
        self.assertEqual(0xAFE, parse_dec_or_hex('0xAFE'))

    def test_cannot_parse_invalid_number(self):
        for string in INVALID_NUMERIC_STRINGS:
            with self.assertRaises(ValueError):
                parse_dec_or_hex(string)

# endregion

# region builtin


INVALID_BUILTIN_TYPE_NAMES = [
    'binary_fixed(2x22)',  # malformed number
    'binary_fixed(0x8Y8)',
    'binary_fixed(x)',
    'uint33',  # unsupported size
    ' uint32',  # invalid spacing
    'uint32 ',
    'FooBar'  # non-builtin
]


class IsBuiltinTest(unittest.TestCase):
    def test_true_for_positives(self):
        for builtin_tuple in BUILTIN_TYPE_TUPLES:
            # Act:
            result = is_builtin(builtin_tuple[0])

            # Assert:
            self.assertTrue(result)

    def test_false_for_negatives(self):
        for string in INVALID_BUILTIN_TYPE_NAMES:
            # Act:
            result = is_builtin(string)

            # Assert:
            self.assertFalse(result)


class ParseBuiltinTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Act:
        result = parse_builtin(line)

        # Assert:
        self.assertEqual(expected_result, result)

    def test_can_parse_int_builtin(self):
        for int_tuple in INT_TYPE_TUPLES:
            # Act + Assert:
            self._assert_parse(
                int_tuple[0],
                {'type': 'byte', 'signedness': 'signed', 'size': int_tuple[1]})

    def test_can_parse_uint_builtin(self):
        for uint_tuple in UINT_TYPE_TUPLES:
            # Act + Assert:
            self._assert_parse(
                uint_tuple[0],
                {'type': 'byte', 'signedness': 'unsigned', 'size': uint_tuple[1]})

    def test_can_parse_binary_fixed_builtin(self):
        for size_tuple in [('32', 32), ('0x20', 32), ('25', 25)]:
            # Act + Assert:
            self._assert_parse(
                'binary_fixed({0})'.format(size_tuple[0]),
                {'type': 'byte', 'signedness': 'unsigned', 'size': size_tuple[1]})

    def test_cannot_parse_invalid_builtin(self):
        # Arrange:
        for type_name in INVALID_BUILTIN_TYPE_NAMES:
            # Act + Assert:
            with self.assertRaises(CatsParseException):
                parse_builtin(type_name)

# endregion



from catbuffer_parser.CatsParseException import CatsParseException


class SingleLineParserTestUtils:
    def __init__(self, parser_factory_type, unittest):
        self.parser_factory_type = parser_factory_type
        self.unittest = unittest

    def assert_parse(self, line, expected_result):
        # Arrange:
        parser = self.parser_factory_type().create()

        # Act:
        result = parser.process_line(line)

        # Assert:
        self.unittest.assertEqual(expected_result, result)

    def assert_parse_exception(self, line, exception_type=CatsParseException):
        # Arrange:
        parser = self.parser_factory_type().create()

        # Sanity:
        self.unittest.assertTrue(self.parser_factory_type().is_match(line))

        # Act + Assert:
        with self.unittest.assertRaises(exception_type):
            parser.process_line(line)

    def assert_parse_exceptions(self, invalid_lines):
        # Arrange:
        for line in invalid_lines:
            # Act + Assert:
            self.assert_parse_exception(line)

    def assert_naming(self, pattern, valid_names, invalid_names):
        for name in valid_names:
            # Act + Assert: no exception
            self.parser_factory_type().create().process_line(pattern.format(name))

        for name in invalid_names:
            # Act + Assert: exception
            self.assert_parse_exception(pattern.format(name))


class MultiLineParserTestUtils(SingleLineParserTestUtils):
    def assert_parse(self, line, expected_result):
        # Arrange:
        parser = self.parser_factory_type().create()

        # Act:
        parser.process_line(line)
        result = parser.commit()

        # Assert:
        self.unittest.assertEqual(expected_result, result)


class ParserFactoryTestUtils:
    def __init__(self, parser_factory_type, unittest):
        self.parser_factory_type = parser_factory_type
        self.unittest = unittest

    def assert_positives(self, matches):
        # Arrange:
        factory = self.parser_factory_type()

        # Act + Assert:
        for line in matches:
            self.unittest.assertTrue(factory.is_match(line))

    def assert_negatives(self, matches):
        # Arrange:
        factory = self.parser_factory_type()

        # Act + Assert:
        for line in matches:
            self.unittest.assertFalse(factory.is_match(line))



from .parserutils import parse_builtin, require_user_type_name
from .RegexParserFactory import RegexParserFactory


class AliasParser:
    """Parser for `using` statements"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)

        # aliases are only supported for builtin types
        return (
            require_user_type_name(match.group(1)),
            parse_builtin(match.group(2))
        )


class AliasParserFactory(RegexParserFactory):
    """Factory for creating alias parsers"""
    def __init__(self):
        super().__init__(r'using (\S+) = (\S+)', AliasParser)



import unittest

from catbuffer_parser.CommentParser import CommentParser


class CommentParserTest(unittest.TestCase):
    def test_try_process_line_returns_true_only_for_comment_lines(self):
        # Arrange:
        parser = CommentParser()

        # Act + Assert:
        for line in ['# foo bar', '#', '# $$$']:
            self.assertTrue(parser.try_process_line(line))

        for line in [' # foo bar', 'foo bar']:
            self.assertFalse(parser.try_process_line(line))

    def test_no_comments_are_present_initially(self):
        # Arrange:
        parser = CommentParser()

        # Act:
        result = parser.commit()

        # Assert:
        self.assertEqual({'comments': ''}, result)

    def test_can_add_single_line_comment(self):
        # Arrange:
        parser = CommentParser()

        # Act:
        parser.try_process_line('# this is a comment')
        result = parser.commit()

        # Assert:
        self.assertEqual({'comments': 'this is a comment'}, result)

    def test_can_add_multi_line_comment(self):
        # Arrange:
        parser = CommentParser()

        # Act:
        parser.try_process_line('# this is a comment')
        parser.try_process_line('# foo bar')
        result = parser.commit()

        # Assert:
        self.assertEqual({'comments': 'this is a comment foo bar'}, result)

    def test_post_processing_removes_leading_and_trailing_whitespace_per_line(self):
        # Arrange:
        parser = CommentParser()

        # Act:
        parser.try_process_line('#    this is a comment   ')
        parser.try_process_line('#      foo   bar   ')
        result = parser.commit()

        # Assert:
        self.assertEqual({'comments': 'this is a comment foo   bar'}, result)

    def test_can_reuse_parser(self):
        # Arrange:
        parser = CommentParser()

        # Act:
        parser.try_process_line('# this is a comment')
        result1 = parser.commit()

        parser.try_process_line('# foo bar')
        result2 = parser.commit()

        # Assert:
        self.assertEqual({'comments': 'this is a comment'}, result1)
        self.assertEqual({'comments': 'foo bar'}, result2)



class CommentParser:
    """Reusable aspect parser for comment statements"""
    def __init__(self):
        self.comments = []

    def try_process_line(self, line):
        """Processes the current line if and only if it is a comment line"""
        if line.startswith('#'):
            self.comments.append(line[1:])
            return True

        return False

    def commit(self):
        """Postprocesses and clears all captured comments"""
        comments = ' '.join(comment.strip() for comment in self.comments)
        self.comments.clear()
        return {'comments': comments}



import unittest

from catbuffer_parser.ImportParser import ImportParserFactory, ImportResult

from .ParserTestUtils import ParserFactoryTestUtils, SingleLineParserTestUtils

VALID_IMPORT_FILE_NAMES = ['A', 'aBzZzac09', 'aa bb.cats', 'foo.cats', 'foo bar', 'foo bar.cats', '$^$']


class ImportParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(ImportParserFactory, self).assert_positives([
            'import "{0}"'.format(import_file) for import_file in VALID_IMPORT_FILE_NAMES
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(ImportParserFactory, self).assert_negatives([
            ' import "A"', 'import "A" ', 'import ""', 'import "aa\taa"', 'foo $$$'
        ])


class ImportParserTest(unittest.TestCase):
    def test_can_parse_import(self):
        for import_file in VALID_IMPORT_FILE_NAMES:
            # Act + Assert:
            SingleLineParserTestUtils(ImportParserFactory, self).assert_parse(
                'import "{0}"'.format(import_file),
                ImportResult(import_file))



from .CatsParseException import CatsParseException
from .CompositeTypeParser import CompositeTypeParser
from .parserutils import parse_builtin, parse_dec_or_hex, require_primitive, require_property_name, require_user_type_name
from .RegexParserFactory import RegexParserFactory


class EnumParser(CompositeTypeParser):
    """Parser for `enum` statements"""
    def __init__(self, regex):
        super().__init__(regex, [EnumValueParserFactory()])

    def process_line(self, line):
        match = self.regex.match(line)
        self.type_name = require_user_type_name(match.group(1))

        base_type = require_primitive(match.group(2))
        builtin_type_descriptor = parse_builtin(base_type)
        self.type_descriptor = {
            'type': 'enum',
            'size': builtin_type_descriptor['size'],
            'signedness': builtin_type_descriptor['signedness'],
            'values': []
        }

    def append(self, property_value_descriptor):
        self._require_unknown_property(property_value_descriptor['name'])

        self.type_descriptor['values'].append(property_value_descriptor)

    def _require_unknown_property(self, property_name):
        if any(property_name == property_type_descriptor['name'] for property_type_descriptor in self.type_descriptor['values']):
            raise CatsParseException('duplicate definition for enum value "{0}"'.format(property_name))


class EnumParserFactory(RegexParserFactory):
    """Factory for creating enum parsers"""
    def __init__(self):
        super().__init__(r'enum (\S+) : (u?int\d+)', EnumParser)


class EnumValueParser:
    """Parser for enum values"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        return {'name': require_property_name(match.group(1)), 'value': parse_dec_or_hex(match.group(2))}


class EnumValueParserFactory(RegexParserFactory):
    """Factory for creating enum value parsers"""
    def __init__(self):
        super().__init__(r'(\S+) = (\S+)', EnumValueParser)



VALID_USER_TYPE_NAMES = ['FooBar', 'Foo', 'FooBarBaz', 'Foo123', 'Foo123BAZ']
INVALID_USER_TYPE_NAMES = ['fooBar', 'foo_bar', 'Foo_Bar', '_fooBar', 'm_fooBar', 'FOO_BAR']

VALID_PROPERTY_NAMES = ['fooBar', 'foo', 'fooBarBaz', 'foo123', 'foo123BAZ', 'foo_Bar', 'm_fooBar']
INVALID_PROPERTY_NAMES = ['FooBar', 'Foo_bar', '_fooBar', 'FOO_BAR']

VALID_PRIMITIVE_NAMES = ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64']
INVALID_PRIMITIVE_NAMES = ['sint8', 'vint32', ' uint8', 'uint16 ', 'uint 32', 'uint63', 'foo']

INT_TYPE_TUPLES = [('int8', 1, 'signed'), ('int16', 2, 'signed'), ('int32', 4, 'signed'), ('int64', 8, 'signed')]
UINT_TYPE_TUPLES = [('uint8', 1, 'unsigned'), ('uint16', 2, 'unsigned'), ('uint32', 4, 'unsigned'), ('uint64', 8, 'unsigned')]
PRIMITIVE_TYPE_TUPLES = INT_TYPE_TUPLES + UINT_TYPE_TUPLES
BUILTIN_TYPE_TUPLES = PRIMITIVE_TYPE_TUPLES + [
    ('binary_fixed(32)', 32, 'unsigned'), ('binary_fixed(0x20)', 32, 'unsigned'), ('binary_fixed(25)', 25, 'unsigned')
]



import unittest

from catbuffer_parser.CatsParseException import CatsParseException
from catbuffer_parser.ScopeManager import ScopeManager


class ScopeManagerTest(unittest.TestCase):
    def test_manager_initially_has_default_scope(self):
        # Arrange:
        manager = ScopeManager()

        # Act:
        scope = manager.scope()

        # Assert:
        self.assertEqual(['<unknown>:0'], scope)

    def test_can_increment_line_number(self):
        # Arrange:
        manager = ScopeManager()

        # Act:
        manager.increment_line_number()
        manager.increment_line_number()
        manager.increment_line_number()
        scope = manager.scope()

        # Assert:
        self.assertEqual(['<unknown>:3'], scope)

    @staticmethod
    def _initialize_manager_with_three_scopes():
        # Arrange:
        manager = ScopeManager()

        # Act:
        manager.push_scope('zeta')
        manager.increment_line_number()
        manager.push_scope('beta')
        manager.increment_line_number()
        manager.increment_line_number()
        manager.push_scope('gamma')
        return manager

    def test_can_push_scopes(self):
        # Arrange:
        manager = self._initialize_manager_with_three_scopes()

        # Act:
        scope = manager.scope()

        # Assert:
        self.assertEqual(['gamma:0', 'beta:2', 'zeta:1', '<unknown>:0'], scope)

    def test_can_pop_some_scopes(self):
        # Arrange:
        manager = self._initialize_manager_with_three_scopes()

        # Act:
        manager.pop_scope()
        manager.pop_scope()
        scope = manager.scope()

        # Assert:
        self.assertEqual(['zeta:1', '<unknown>:0'], scope)

    def test_can_pop_all_scopes(self):
        # Arrange:
        manager = self._initialize_manager_with_three_scopes()

        # Act:
        manager.pop_scope()
        manager.pop_scope()
        manager.pop_scope()
        scope = manager.scope()

        # Assert:
        self.assertEqual(['<unknown>:0'], scope)

    def test_cannot_pop_default_scope(self):
        # Arrange:
        manager = ScopeManager()
        manager.push_scope('zeta')
        manager.pop_scope()

        # Act + Assert
        with self.assertRaises(CatsParseException):
            manager.pop_scope()



import unittest

from catbuffer_parser.CatsParser import CatsParseException, CatsParser


def parse_all(lines, imports=None):
    # Act:
    if imports is None:
        imports = []

    parser = CatsParser(imports.append)
    for line in lines:
        parser.process_line(line)

    return parser.type_descriptors()


def uint_descriptor(size, explicit_type=True):
    descriptor = {'signedness': 'unsigned', 'size': size}
    if explicit_type:
        descriptor['type'] = 'byte'

    return descriptor


def int_descriptor(size, explicit_type=True):
    descriptor = uint_descriptor(size, explicit_type)
    descriptor['signedness'] = 'signed'
    return descriptor


class CatsParserTests(unittest.TestCase):
    # pylint: disable=too-many-public-methods

    # region utils

    def _assert_parse_delayed_exception(self, lines):
        # Arrange:
        imports = []
        parser = CatsParser(imports.append)
        for line in lines[:-1]:
            parser.process_line(line)

        # Act + Assert:
        with self.assertRaises(CatsParseException):
            parser.process_line(lines[-1])

    def _assert_parse_commit_exception(self, lines):
        # Arrange:
        imports = []
        parser = CatsParser(imports.append)
        for line in lines:
            parser.process_line(line)

        # Act + Assert:
        with self.assertRaises(CatsParseException):
            parser.commit()

    # endregion

    # region empty + basic

    def test_no_types_are_exposed_initially(self):
        # Act:
        parser = CatsParser(None)

        # Assert:
        self.assertEqual(0, len(parser.type_descriptors()))

    def test_no_types_are_extracted_from_blank_lines(self):
        # Act:
        type_descriptors = parse_all([
            '',
            '',
            ''
        ])

        # Assert:
        self.assertEqual(0, len(type_descriptors))

    def test_parse_is_aborted_on_unknown_line(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'using Foo = uint8',
            'alias Bar = uint8'
        ])

    # endregion

    # region comments

    def test_unattached_comments_are_ignored(self):
        # Act:
        type_descriptors = parse_all([
            '# comment alias',
            '# another comment',
            '',  # should clear previous comments
            'using MosaicId = uint64'
        ])

        # Assert:
        self.assertEqual(1, len(type_descriptors))
        self.assertEqual(type_descriptors['MosaicId'], {**uint_descriptor(8), 'comments': ''})

    def test_previously_attached_comments_are_ignored(self):
        # Act:
        type_descriptors = parse_all([
            '# comment one',
            'using Age = uint64',
            '# comment two',
            'using Year = uint16'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Age'], {**uint_descriptor(8), 'comments': 'comment one'})
        self.assertEqual(type_descriptors['Year'], {**uint_descriptor(2), 'comments': 'comment two'})

    # endregion

    # region imports

    def test_can_parse_valid_import(self):
        # Act:
        imports = []
        type_descriptors = parse_all(['import "foo.cats"'], imports)

        # Assert:
        self.assertEqual(0, len(type_descriptors))
        self.assertEqual(['foo.cats'], imports)

    # endregion

    # region alias

    def test_can_parse_alias(self):
        # Act:
        type_descriptors = parse_all([
            'using MosaicId = uint64',
            '# unique account identifier',
            'using Address = binary_fixed(25)'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['MosaicId'], {**uint_descriptor(8), 'comments': ''})
        self.assertEqual(type_descriptors['Address'], {**uint_descriptor(25), 'comments': 'unique account identifier'})

    # endregion

    # region struct

    def test_can_parse_struct_builtin_types(self):
        # Act:
        type_descriptors = parse_all([
            '# binary layout for a pair',
            'struct Pair',
            '\t# some field comment',
            '\tfooBar = uint64',
            '\tbaz = binary_fixed(25)'
        ])

        # Assert:
        self.assertEqual(1, len(type_descriptors))
        self.assertEqual(type_descriptors['Pair'], {'type': 'struct', 'comments': 'binary layout for a pair', 'layout': [
            {'name': 'fooBar', **uint_descriptor(8), 'comments': 'some field comment'},
            {'name': 'baz', **uint_descriptor(25), 'comments': ''}
        ]})

    def test_can_parse_struct_custom_types(self):
        # Act:
        type_descriptors = parse_all([
            'using MosaicId = uint16',
            'using Amount = uint16',
            '# binary layout for a mosaic',
            'struct Mosaic',
            '\t# mosaic identifier',
            '\tmosaicId = MosaicId',
            '\tamount = Amount'
        ])

        # Assert:
        self.assertEqual(3, len(type_descriptors))
        self.assertEqual(type_descriptors['Mosaic'], {'type': 'struct', 'comments': 'binary layout for a mosaic', 'layout': [
            {'name': 'mosaicId', 'type': 'MosaicId', 'comments': 'mosaic identifier'},
            {'name': 'amount', 'type': 'Amount', 'comments': ''}
        ]})

    def test_can_parse_struct_conditional_types(self):
        # Act:
        type_descriptors = parse_all([
            'enum Shape : uint8',
            '\tcircle = 4',
            '\trectangle = 9',
            'using Circ = uint16',
            'using Perm = uint16',
            'struct Enclosing',
            '\tdiscriminator = Shape',
            '\t# u part 1',
            '\tcircumference = Circ if discriminator has circle',
            '\t# union pt 2',
            '\tperimiter = Perm if discriminator equals rectangle'
        ])

        # Assert:
        self.assertEqual(4, len(type_descriptors))
        self.assertEqual(type_descriptors['Enclosing'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'discriminator', 'type': 'Shape', 'comments': ''},
            {
                'name': 'circumference', 'type': 'Circ',
                'condition': 'discriminator', 'condition_value': 'circle', 'condition_operation': 'has',
                'comments': 'u part 1'
            },
            {
                'name': 'perimiter', 'type': 'Perm',
                'condition': 'discriminator', 'condition_value': 'rectangle', 'condition_operation': 'equals',
                'comments': 'union pt 2'
            }
        ]})

    def test_can_parse_struct_conditional_types_with_inline_member(self):
        # Act:
        type_descriptors = parse_all([
            'enum Shape : uint8',
            '\tcircle = 4',
            '\trectangle = 9',
            'using Circ = uint16',
            'using Perm = uint16',
            'struct Version',
            '\tversion = uint16',
            'struct Enclosing',
            '\tinline Version',
            '\tdiscriminator = Shape',
            '\t# u part 1',
            '\tcircumference = Circ if discriminator has circle',
            '\t# union pt 2',
            '\tperimiter = Perm if discriminator equals rectangle'
        ])

        # Assert:
        self.assertEqual(5, len(type_descriptors))
        self.assertEqual(type_descriptors['Enclosing'], {'type': 'struct', 'comments': '', 'layout': [
            {'type': 'Version', 'disposition': 'inline', 'comments': ''},
            {'name': 'discriminator', 'type': 'Shape', 'comments': ''},
            {
                'name': 'circumference', 'type': 'Circ',
                'condition': 'discriminator', 'condition_value': 'circle', 'condition_operation': 'has',
                'comments': 'u part 1'
            },
            {
                'name': 'perimiter', 'type': 'Perm',
                'condition': 'discriminator', 'condition_value': 'rectangle', 'condition_operation': 'equals',
                'comments': 'union pt 2'
            }
        ]})

    def test_can_parse_struct_conditional_types_trailing_discriminator(self):
        # Act:
        type_descriptors = parse_all([
            'enum Shape : uint8',
            '\tcircle = 4',
            '\trectangle = 9',
            'using Circ = uint16',
            'using Perm = uint16',
            'struct Enclosing',
            '\t# u part 1',
            '\tcircumference = Circ if discriminator has circle',
            '\t# union pt 2',
            '\tperimiter = Perm if discriminator equals rectangle',
            '\tdiscriminator = Shape'
        ])

        # Assert:
        self.assertEqual(4, len(type_descriptors))
        self.assertEqual(type_descriptors['Enclosing'], {'type': 'struct', 'comments': '', 'layout': [
            {
                'name': 'circumference', 'type': 'Circ',
                'condition': 'discriminator', 'condition_value': 'circle', 'condition_operation': 'has',
                'comments': 'u part 1'
            },
            {
                'name': 'perimiter', 'type': 'Perm',
                'condition': 'discriminator', 'condition_value': 'rectangle', 'condition_operation': 'equals',
                'comments': 'union pt 2'
            },
            {'name': 'discriminator', 'type': 'Shape', 'comments': ''}
        ]})

    def test_can_parse_struct_array_types(self):
        # Act:
        type_descriptors = parse_all([
            'using Truck = uint16',
            'using Car = uint16',
            'struct Fleet',
            '\tcarCount = uint8',
            '# all trucks in the fleet',
            '\ttrucks = array(Truck, 10)',
            '\tcars = array(Car, carCount)'
        ])

        # Assert:
        self.assertEqual(3, len(type_descriptors))
        self.assertEqual(type_descriptors['Fleet'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'carCount', **uint_descriptor(1), 'comments': ''},
            {'name': 'trucks', 'type': 'Truck', 'size': 10, 'comments': 'all trucks in the fleet'},
            {'name': 'cars', 'type': 'Car', 'size': 'carCount', 'comments': ''}
        ]})

    def test_can_parse_struct_sorted_array_types(self):
        # Act:
        type_descriptors = parse_all([
            'struct Face',
            '\teyeColor = uint8',
            'struct Tracking',
            '\tfaces = array(Face, 10, sort_key=eyeColor)'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Tracking'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'faces', 'type': 'Face', 'size': 10, 'sort_key': 'eyeColor', 'comments': ''}
        ]})

    def test_can_parse_struct_vararray_types(self):
        # Act:
        type_descriptors = parse_all([
            'using Car = uint16',
            'struct Fleet',
            '\tcarsSize = uint8',
            '# all cars in the fleet',
            '\tcars = array(Car, size=carsSize)'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Fleet'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'carsSize', **uint_descriptor(1), 'comments': ''},
            {'name': 'cars', 'type': 'Car', 'size': 'carsSize', 'disposition': 'var', 'comments': 'all cars in the fleet'}
        ]})

    def test_can_parse_struct_array_fill_types(self):
        # Act:
        type_descriptors = parse_all([
            'using Car = uint16',
            'struct Fleet',
            '# all cars in the fleet',
            '\tcars = array(Car, __FILL__)'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Fleet'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'cars', 'type': 'Car', 'size': 0, 'disposition': 'fill', 'comments': 'all cars in the fleet'}
        ]})

    def test_cannot_parse_struct_vararray_numeric_size_types(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'using Car = uint16',
            'struct Fleet',
            '# all cars in the fleet',
            '\tcars = array(Car, size=123)'
        ])

    def test_cannot_parse_struct_vararray_fill_types(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'using Car = uint16',
            'struct Fleet',
            '# all cars in the fleet',
            '\tcars = array(Car, size=__FILL__)'
        ])

    def test_can_parse_struct_closed_by_other_type(self):
        # Act:
        type_descriptors = parse_all([
            'struct Mosaic',
            '\tmosaicId = uint64',
            '',
            '\tamount = uint32',
            'using Address = binary_fixed(25)'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Mosaic'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'mosaicId', **uint_descriptor(8), 'comments': ''},
            {'name': 'amount', **uint_descriptor(4), 'comments': ''}
        ]})

    def test_can_parse_struct_with_inline_member(self):
        # Act:
        type_descriptors = parse_all([
            'struct Placeholder',
            'struct Pair',
            '\tfooBar = uint64',
            '# some placeholder comment',
            '\tinline Placeholder',
            '\tbaz = uint32'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Pair'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'fooBar', **uint_descriptor(8), 'comments': ''},
            {'type': 'Placeholder', 'disposition': 'inline', 'comments': 'some placeholder comment'},
            {'name': 'baz', **uint_descriptor(4), 'comments': ''}
        ]})

    def test_can_parse_struct_with_const_member(self):
        # Act:
        type_descriptors = parse_all([
            'struct Pair',
            '\tfooBar = uint64',
            '# some const comment',
            '\tconst int8 tupleSize = 2',
            '\tbaz = uint32'
        ])

        # Assert:
        self.assertEqual(1, len(type_descriptors))
        self.assertEqual(type_descriptors['Pair'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'fooBar', **uint_descriptor(8), 'comments': ''},
            {'name': 'tupleSize', **int_descriptor(1), 'disposition': 'const', 'value': 2, 'comments': 'some const comment'},
            {'name': 'baz', **uint_descriptor(4), 'comments': ''}
        ]})

    def test_cannot_parse_struct_with_unknown_member_type(self):
        # Act + Assert:
        for type_name in ['MosaicId', 'array(MosaicId, 10)']:
            self._assert_parse_delayed_exception([
                'struct Foo',
                '\tid = {0}'.format(type_name)
            ])

    def test_cannot_parse_struct_with_non_enum_condition_link(self):
        # Act + Assert:
        self._assert_parse_commit_exception([
            'using Shape = uint8',
            'using Circ = uint16',
            'struct Enclosing',
            '\tdiscriminator = Shape',
            '\tcircumference = Circ if discriminator equals 123'
        ])

    def test_cannot_parse_struct_with_unknown_condition_value(self):
        # Act + Assert:
        self._assert_parse_commit_exception([
            'enum Shape : uint8',
            '\tcircle = 1',
            'using Circ = uint16',
            'struct Enclosing',
            '\tdiscriminator = Shape',
            '\tcircumference = Circ if discriminator equals hexagon'
        ])

    def test_cannot_parse_struct_with_unknown_array_size(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'using MosaicId = uint16',
            'struct Foo',
            '\tids = array(MosaicId, numMosaics)'
        ])

    def test_cannot_parse_struct_with_unknown_sort_key(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'using Face = uint16',
            'struct Tracking',
            '\tfaces = array(Face, 10, sort_key=eyeColor)'
        ])

    def test_cannot_parse_struct_with_unknown_inline_type(self):
        # Act + Assert:
        for type_name in ['MosaicId', 'array(MosaicId, 10)', 'uint8', 'binary_fixed(25)']:
            self._assert_parse_delayed_exception([
                'struct Foo',
                '\tinline {0}'.format(type_name)
            ])

    def test_cannot_parse_struct_with_unknown_const_type(self):
        # Act + Assert:
        for type_name in ['uint7', 'binary_fixed(25)', 'Car']:
            self._assert_parse_delayed_exception([
                'struct Foo',
                '\tconst {0} bar = 123'.format(type_name)
            ])

    # endregion

    # region enum

    def test_can_parse_enum_values(self):
        # Act:
        type_descriptors = parse_all([
            '# enumeration of entity types',
            'enum EntityType : uint16',
            '\t# transfer transaction type',
            '\ttransfer = 7',
            '\thashLock = 0x0C'
        ])

        # Assert:
        self.assertEqual(1, len(type_descriptors))
        self.assertEqual(type_descriptors['EntityType'], {
            'type': 'enum', **uint_descriptor(2, False), 'comments': 'enumeration of entity types', 'values': [
                {'name': 'transfer', 'value': 7, 'comments': 'transfer transaction type'},
                {'name': 'hashLock', 'value': 12, 'comments': ''}
            ]
        })

    def test_can_parse_enum_closed_by_other_type(self):
        # Act:
        type_descriptors = parse_all([
            'enum EntityType : uint16',
            '\ttransfer = 7',
            '',
            '\thashLock = 0x0C',
            'using Address = binary_fixed(25)'
        ])

        # Assert:
        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['EntityType'], {'type': 'enum', **uint_descriptor(2, False), 'comments': '', 'values': [
            {'name': 'transfer', 'value': 7, 'comments': ''},
            {'name': 'hashLock', 'value': 12, 'comments': ''}
        ]})

    def test_cannot_parse_enum_with_non_numeric_value(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'enum EntityType : uint16',
            '\ttransfer = QZ'
        ])

    # endregion

    # region scoping

    def test_cannot_parse_as_outer_scope_from_inner_scope(self):
        # Arrange:
        valid_headers = [
            ['struct Mosaic', '\tmosaicId = uint64'],
            ['enum EntityType : uint16', '\ttransfer = 7']
        ]

        # Act + Assert:
        # - using is not valid in composite scope
        for header in valid_headers:
            self._assert_parse_delayed_exception(header + ['\tusing Address = binary_fixed(25)'])

    def test_cannot_parse_schema_with_duplicate_type_names(self):
        # Arrange:
        valid_declarations = [
            ['struct Bar', '\tfoo = uint64'],
            ['enum Bar : uint16', '\tbaz = 7'],
            ['using Bar = uint16']
        ]

        # Act + Assert:
        counter = 0
        for declaration1 in valid_declarations:
            for declaration2 in valid_declarations:
                lines = declaration1 + declaration2
                if not declaration2[0].startswith('using'):
                    lines += ['using Baz = uint32']  # make sure all composites are closed

                self._assert_parse_delayed_exception(lines)
                counter += 1

        # Sanity:
        self.assertEqual(9, counter)

    def test_cannot_parse_schema_with_duplicate_struct_property_names_in_same_scope(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'struct Bar',
            '\tfoo = uint8',
            '\tfoo = uint16'
        ])

    def test_cannot_parse_schema_with_duplicate_enum_property_names_in_same_scope(self):
        # Act + Assert:
        self._assert_parse_delayed_exception([
            'enum Bar : uint16',
            '\tfoo = 4',
            '\tfoo = 9'
        ])

    def test_can_parse_schema_with_duplicate_struct_property_names_in_different_scopes(self):
        # Act:
        type_descriptors = parse_all([
            'struct Bar',
            '\tfoo = uint8',
            '',
            'struct Baz',
            '\tfoo = uint16'
        ])

        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Bar'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'foo', **uint_descriptor(1), 'comments': ''}
        ]})
        self.assertEqual(type_descriptors['Baz'], {'type': 'struct', 'comments': '', 'layout': [
            {'name': 'foo', **uint_descriptor(2), 'comments': ''}
        ]})

    def test_can_parse_schema_with_duplicate_enum_property_names_in_different_scopes(self):
        # Act:
        type_descriptors = parse_all([
            'enum Bar : uint16',
            '\tfoo = 4',
            '',
            'enum Baz : int32',
            '\tfoo = 9'
        ])

        self.assertEqual(2, len(type_descriptors))
        self.assertEqual(type_descriptors['Bar'], {'type': 'enum', **uint_descriptor(2, False), 'comments': '', 'values': [
            {'name': 'foo', 'value': 4, 'comments': ''}
        ]})
        self.assertEqual(type_descriptors['Baz'], {'type': 'enum', **int_descriptor(4, False), 'comments': '', 'values': [
            {'name': 'foo', 'value': 9, 'comments': ''}
        ]})

    # endregion

    # region ordering

    def test_type_definition_order_is_preserved(self):
        # Act:
        type_descriptors = parse_all([
            'using Truck = uint16',
            'struct Fleet',
            '\tcarCount = uint8',
            'enum Bar : uint16',
            '\tfoo = 4',
            'using Car = uint16'
        ])

        # Assert:
        self.assertEqual(4, len(type_descriptors))
        self.assertEqual(list(type_descriptors.keys()), ['Truck', 'Fleet', 'Bar', 'Car'])

    # endregion



import unittest

from catbuffer_parser.CatsParseException import CatsParseException
from catbuffer_parser.StructParser import (StructArrayMemberParserFactory, StructConstParserFactory, StructInlineParserFactory,
                                           StructParserFactory, StructScalarMemberParserFactory)

from .constants import (BUILTIN_TYPE_TUPLES, INVALID_PROPERTY_NAMES, INVALID_USER_TYPE_NAMES, VALID_PRIMITIVE_NAMES, VALID_PROPERTY_NAMES,
                        VALID_USER_TYPE_NAMES)
from .ParserTestUtils import MultiLineParserTestUtils, ParserFactoryTestUtils, SingleLineParserTestUtils

# region StructParserTest


class StructParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(StructParserFactory, self).assert_positives([
            'struct F', 'struct Foo', 'struct FooZA09za', 'struct foo', 'struct 8oo', 'struct $^^$'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(StructParserFactory, self).assert_negatives([
            ' struct Foo', 'struct Foo ', 'struct ', 'struct foo bar'
        ])


class StructParserTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Assert:
        MultiLineParserTestUtils(StructParserFactory, self).assert_parse(line, expected_result)

    def _assert_parse_exception(self, line):
        # Assert:
        MultiLineParserTestUtils(StructParserFactory, self).assert_parse_exception(line)

    def test_parser_exposes_custom_factories(self):
        # Act
        parser = StructParserFactory().create()

        # Assert
        self.assertEqual(4, len(parser.factories()))

    def test_can_parse_type_declaration(self):
        # Act + Assert:
        self._assert_parse(
            'struct Car',
            ('Car', {'type': 'struct', 'layout': []}))

    def test_struct_names_must_have_type_name_semantics(self):
        # Assert:
        MultiLineParserTestUtils(StructParserFactory, self).assert_naming('struct {0}', VALID_USER_TYPE_NAMES, INVALID_USER_TYPE_NAMES)

    def test_can_append_scalar(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [{'name': 'foo'}, {'name': 'bar'}]}), result)

    def test_can_append_scalar_conditional(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar', 'condition': 'foo', 'condition_value': 'red'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [
            {'name': 'foo'},
            {'name': 'bar', 'condition': 'foo', 'condition_value': 'red'}
        ]}), result)

    def test_can_append_scalar_conditional_trailing_discriminator(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'bar', 'condition': 'foo', 'condition_value': 'red'})
        parser.append({'name': 'foo'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [
            {'name': 'bar', 'condition': 'foo', 'condition_value': 'red'},
            {'name': 'foo'}
        ]}), result)

    def test_cannot_append_scalar_with_invalid_condition_reference(self):
        for condition in ['baz', '10']:
            # Arrange:
            parser = StructParserFactory().create()

            # Act:
            parser.process_line('struct Car')
            parser.append({'name': 'foo'})
            parser.append({'name': 'bar', 'condition': condition, 'condition_value': 'red'})

            # Assert:
            with self.assertRaises(CatsParseException):
                parser.commit()

    def test_can_append_array_with_numeric_size(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar', 'size': 10})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [{'name': 'foo'}, {'name': 'bar', 'size': 10}]}), result)

    def test_can_append_array_with_valid_size_reference(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar', 'size': 'foo'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [{'name': 'foo'}, {'name': 'bar', 'size': 'foo'}]}), result)

    def test_can_append_array_with_valid_size_reference_and_inline(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'disposition': 'inline', 'type': 'Vehicle'})
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar', 'size': 'foo'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [
            {'disposition': 'inline', 'type': 'Vehicle'},
            {'name': 'foo'},
            {'name': 'bar', 'size': 'foo'}
        ]}), result)

    def test_cannot_append_array_with_invalid_size_reference(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})

        # Assert:
        with self.assertRaises(CatsParseException):
            parser.append({'name': 'bar', 'size': 'fob'})

    def test_can_append_multiple_properties_with_same_name_and_different_disposition(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'foo', 'disposition': 'const'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [{'name': 'foo'}, {'name': 'foo', 'disposition': 'const'}]}), result)

    def test_cannot_append_multiple_properties_with_same_name_and_disposition(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'bar'})

        # Assert:
        with self.assertRaises(CatsParseException):
            parser.append({'name': 'foo'})

    def test_can_append_unnamed_descriptor(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'disposition': 'inline', 'type': 'Foo'})
        parser.append({'name': 'foo'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [
            {'disposition': 'inline', 'type': 'Foo'},
            {'name': 'foo'}
        ]}), result)

    def test_can_append_array_with_fill_disposition_last(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'cat', 'disposition': 'fill'})
        result = parser.commit()

        # Assert:
        self.assertEqual(('Car', {'type': 'struct', 'layout': [
            {'name': 'foo'},
            {'name': 'cat', 'disposition': 'fill'}
        ]}), result)

    def test_cannot_append_property_after_array_with_fill_disposition(self):
        # Arrange:
        parser = StructParserFactory().create()

        # Act:
        parser.process_line('struct Car')
        parser.append({'name': 'foo'})
        parser.append({'name': 'cat', 'disposition': 'fill'})

        # Assert:
        with self.assertRaises(CatsParseException):
            parser.append({'name': 'bar'})

# endregion

# region StructConstParserTest


class StructConstParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(StructConstParserFactory, self).assert_positives([
            'const Foo foo = Bar', 'const FOO FOO = BAR', 'const Za09Za FzaZa09 = fzaZa09', 'const !!! $$$ = ###'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(StructConstParserFactory, self).assert_negatives([
            ' const Foo foo = Bar', 'const Foo foo = Bar ', 'const Foo foo =', 'const Foo foo Bar', 'const Foo = Bar', 'const Foo'
        ])


class StructConstParserTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Assert:
        SingleLineParserTestUtils(StructConstParserFactory, self).assert_parse(line, expected_result)

    def _assert_parse_exception(self, line):
        # Assert:
        SingleLineParserTestUtils(StructConstParserFactory, self).assert_parse_exception(line)

    def test_can_parse_uint_type_constant(self):
        # Act + Assert:
        for value in [32, 0x20]:
            self._assert_parse(
                'const uint16 foo = {0}'.format(value),
                {'name': 'foo', 'type': 'byte', 'signedness': 'unsigned', 'size': 2, 'value': 32, 'disposition': 'const'})

    def test_can_parse_custom_type_constant(self):
        # Act + Assert:
        for value in [33, 0x21]:
            self._assert_parse(
                'const ColorShade red = {0}'.format(value),
                {'name': 'red', 'type': 'ColorShade', 'value': 33, 'disposition': 'const'})

    def test_cannot_parse_non_numeric_value(self):
        # Act + Assert:
        for value in ['FOO', 'AF']:
            SingleLineParserTestUtils(StructConstParserFactory, self).assert_parse_exception(
                'const uint16 foo = {0}'.format(value),
                ValueError)

    def test_cannot_parse_binary_fixed_type_constant(self):
        # Act + Assert:
        SingleLineParserTestUtils(StructConstParserFactory, self).assert_parse_exception('const binary_fixed(25) foo = 123')

    def test_member_names_must_have_property_name_semantics(self):
        # Assert:
        SingleLineParserTestUtils(StructConstParserFactory, self).assert_naming(
            'const uint32 {0} = 123',
            VALID_PROPERTY_NAMES,
            INVALID_PROPERTY_NAMES)

    def test_type_names_must_have_type_name_or_uint_semantics(self):
        # Assert:
        SingleLineParserTestUtils(StructConstParserFactory, self).assert_naming(
            'const {0} foo = 123',
            VALID_USER_TYPE_NAMES + VALID_PRIMITIVE_NAMES,
            INVALID_USER_TYPE_NAMES + ['binary_fixed(32)'])

# endregion

# region StructInlineParserTest


class StructInlineParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(StructInlineParserFactory, self).assert_positives([
            'inline Bar', 'inline BAR', 'inline fzaZa09', 'inline $$$'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(StructInlineParserFactory, self).assert_negatives([
            ' inline Bar', 'inline Bar ', 'inline ', ' Bar'
        ])


class StructInlineParserTest(unittest.TestCase):
    def test_can_parse_simple_custom_declaration(self):
        # Act + Assert:
        SingleLineParserTestUtils(StructInlineParserFactory, self).assert_parse(
            'inline Vehicle_',
            {'type': 'Vehicle_', 'disposition': 'inline'})

# endregion

# region StructScalarMemberParser


class StructScalarParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(StructScalarMemberParserFactory, self).assert_positives([
            'foo = bar', 'foo = BAR', 'fzaZa09 = d', '& = $$$', 'foo = fazFZA90',
            'foo = bar if abc equals def', 'foo = bar if abc has def'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(StructScalarMemberParserFactory, self).assert_negatives([
            ' foo = bar', 'foo = bar ', 'foo = ', '= bar', 'foo = array(bar, baz)', 'foo = bar if abc mask def',
            'foo = bar if abc equals', 'foo = bar abc equals def',
            'foo = bar if abc has', 'foo = bar abc has def'
        ])


class StructScalarParserTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Assert:
        SingleLineParserTestUtils(StructScalarMemberParserFactory, self).assert_parse(line, expected_result)

    def test_can_parse_simple_custom_declaration(self):
        # Act + Assert:
        self._assert_parse(
            'car = Vehicle_',
            {'name': 'car', 'type': 'Vehicle_'})

    def test_can_parse_simple_builtin_declaration(self):
        for builtin_tuple in BUILTIN_TYPE_TUPLES:
            # Act + Assert:
            self._assert_parse(
                'car = {0}'.format(builtin_tuple[0]),
                {'name': 'car', 'type': 'byte', 'signedness': builtin_tuple[2], 'size': builtin_tuple[1]})

    def test_can_parse_conditional_custom_declaration_equals(self):
        # Act + Assert:
        self._assert_parse(
            'roadGrade = RoadGrade_ if terrain equals road',
            {'name': 'roadGrade', 'type': 'RoadGrade_', 'condition': 'terrain', 'condition_operation': 'equals', 'condition_value': 'road'})

    def test_can_parse_conditional_custom_declaration_has(self):
        # Act + Assert:
        self._assert_parse(
            'roadGrade = RoadGrade_ if terrain has road',
            {'name': 'roadGrade', 'type': 'RoadGrade_', 'condition': 'terrain', 'condition_operation': 'has', 'condition_value': 'road'})

    def test_member_names_must_have_property_name_semantics(self):
        # Assert:
        SingleLineParserTestUtils(StructScalarMemberParserFactory, self).assert_naming(
            '{0} = uint32',
            VALID_PROPERTY_NAMES,
            INVALID_PROPERTY_NAMES)

# endregion

# region StructArrayMemberParser


VALID_ARRAY_PATTERNS = ['foo = {0}(bar, {1}baz)', '$$$ = {0}(&, {1}**)', '$$$ = {0}(&, {1}**, sort_key=@@)']
INVALID_ARRAY_PATTERNS = [
    ' foo = {0}(bar, {1}baz)', 'foo = {0}(bar, {1}baz) ', 'foo = ', '= {0}(bar, {1}baz)',
    'foo = {0}(bar, {1}baz', 'foo = {0}(bar, {1}baz) if abc equals def', 'foo = {0}(bar, {1}baz) if abc has def'
]
ARRAY_DIMENSION_QUALIFIERS = ['', 'size=']


class StructArrayMemberParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        for dimension_qualifier in ARRAY_DIMENSION_QUALIFIERS:
            ParserFactoryTestUtils(StructArrayMemberParserFactory, self).assert_positives([
                pattern.format('array', dimension_qualifier) for pattern in VALID_ARRAY_PATTERNS
            ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        for dimension_qualifier in ARRAY_DIMENSION_QUALIFIERS:
            ParserFactoryTestUtils(StructArrayMemberParserFactory, self).assert_negatives([
                pattern.format('array', dimension_qualifier) for pattern in INVALID_ARRAY_PATTERNS
            ])


class StructArrayMemberParserTest(unittest.TestCase):
    def _assert_parse(self, line, expected_result):
        # Assert:
        SingleLineParserTestUtils(StructArrayMemberParserFactory, self).assert_parse(line, expected_result)

    def test_can_parse_array_with_non_numeric_size(self):
        for type_name in ['byte', 'Car']:
            # Act + Assert:
            self._assert_parse(
                'vehicles = array({0}, garageSize)'.format(type_name),
                {'name': 'vehicles', 'type': type_name, 'size': 'garageSize'})

    def test_can_parse_array_with_numeric_size(self):
        for type_name in ['byte', 'Car']:
            for numeric_str in ['10', '0x0A']:
                # Act + Assert:
                self._assert_parse(
                    'vehicles = array({0}, {1})'.format(type_name, numeric_str),
                    {'name': 'vehicles', 'type': type_name, 'size': 10})

    def test_can_parse_array_with_fill_size(self):
        for type_name in ['byte', 'Car']:
            # Act + Assert:
            self._assert_parse(
                'vehicles = array({0}, __FILL__)'.format(type_name),
                {'name': 'vehicles', 'type': type_name, 'size': 0, 'disposition': 'fill'})

    def test_can_parse_array_with_sort_key(self):
        # Act + Assert:
        self._assert_parse(
            'vehicles = array(Car, 10, sort_key=bar)',
            {'name': 'vehicles', 'type': 'Car', 'size': 10, 'sort_key': 'bar'})

    def test_can_parse_vararray_with_non_numeric_size(self):
        for type_name in ['byte', 'Car']:
            # Act + Assert:
            self._assert_parse(
                'vehicles = array({0}, size=garageSize)'.format(type_name),
                {'name': 'vehicles', 'type': type_name, 'size': 'garageSize', 'disposition': 'var'})

    def test_can_parse_vararray_with_unsupported_numeric_size(self):
        for type_name in ['byte', 'Car']:
            # Act + Assert: size is not converted for var array
            self._assert_parse(
                'vehicles = array({0}, size=0x0A)'.format(type_name),
                {'name': 'vehicles', 'type': type_name, 'size': '0x0A', 'disposition': 'var'})

    def test_can_parse_vararray_with_unsupported_fill_size(self):
        for type_name in ['byte', 'Car']:
            # Act + Assert: size is not converted for var array
            self._assert_parse(
                'vehicles = array({0}, size=__FILL__)'.format(type_name),
                {'name': 'vehicles', 'type': type_name, 'size': '__FILL__', 'disposition': 'var'})

    def test_can_parse_vararray_with_sort_key(self):
        # Act + Assert:
        self._assert_parse(
            'vehicles = array(Car, size=garageSize, sort_key=bar)',
            {'name': 'vehicles', 'type': 'Car', 'size': 'garageSize', 'disposition': 'var', 'sort_key': 'bar'})

    def test_member_names_must_have_property_name_semantics(self):
        # Assert:
        SingleLineParserTestUtils(StructArrayMemberParserFactory, self).assert_naming(
            '{0} = array(Car, 10)',
            VALID_PROPERTY_NAMES,
            INVALID_PROPERTY_NAMES)

# endregion



import argparse
import os
import pprint

from .CatsParser import CatsParser

try:
    from generators.All import AVAILABLE_GENERATORS
except ImportError:
    AVAILABLE_GENERATORS = {}


class MultiFileParser:
    """CATS parser that resolves imports in global namespace"""
    def __init__(self):
        self.cats_parser = CatsParser(self._process_import_file)
        self.dirname = None
        self.imported_filenames = []

    def set_include_path(self, include_path):
        self.dirname = include_path

    def parse(self, schema_filename):
        self._process_file(schema_filename)

    def _process_import_file(self, filename):
        if filename in self.imported_filenames:
            return

        self.imported_filenames.append(filename)
        filename = os.path.join(self.dirname, filename)
        self._process_file(filename)

    def _process_file(self, filename):
        self.cats_parser.push_scope(filename)

        with open(filename) as input_file:
            lines = input_file.readlines()
            for line in lines:
                self.cats_parser.process_line(line)

        self.cats_parser.commit()
        self.cats_parser.pop_scope()


def _generate_output(generator_name, output_path, schema, options):
    generator_class = AVAILABLE_GENERATORS[generator_name]
    os.makedirs(output_path, exist_ok=True)
    generator = generator_class(schema, options)
    for generated_descriptor in generator:
        output_filename = os.path.join(output_path, generated_descriptor.filename)
        with open(output_filename, 'w', newline='\n') as output_file:
            for line in generated_descriptor.code:
                output_file.write('%s\n' % line)


def main():
    parser = argparse.ArgumentParser(
        prog=None if globals().get('__spec__') is None else 'python -m {}'.format(__spec__.name.partition('.')[0]),
        description='CATS code generator'
    )
    parser.add_argument('-s', '--schema', help='input CATS file', required=True)
    parser.add_argument('-o', '--output', help='output directory, if not provided, _generated/{generator} is used')
    parser.add_argument('-i', '--include', help='schema root directory', default='./schemas')

    generators_list = list(AVAILABLE_GENERATORS.keys())
    parser.add_argument('-g', '--generator', help='generator to use to produce output files', choices=generators_list)
    parser.add_argument('-c', '--copyright', help='file containing copyright data to use with output files', default='../HEADER.inc')
    args = parser.parse_args()

    file_parser = MultiFileParser()
    file_parser.set_include_path(args.include)
    file_parser.parse(args.schema)

    # console output the parsed schema
    printer = pprint.PrettyPrinter(width=140)
    printer.pprint('*** *** ***')
    type_descriptors = file_parser.cats_parser.type_descriptors()
    for key in type_descriptors:
        printer.pprint((key, type_descriptors[key]))

    # generate and output code
    if args.generator:
        output_path = args.output
        if output_path is None:
            output_path = os.path.join('_generated', args.generator)

        _generate_output(args.generator, output_path, type_descriptors, {'copyright': args.copyright})


if '__main__' == __name__:
    main()



import unittest

from catbuffer_parser.AliasParser import AliasParserFactory

from .constants import BUILTIN_TYPE_TUPLES, INVALID_USER_TYPE_NAMES, VALID_USER_TYPE_NAMES
from .ParserTestUtils import ParserFactoryTestUtils, SingleLineParserTestUtils


class AliasParserFactoryTest(unittest.TestCase):
    def test_is_match_returns_true_for_positives(self):
        # Assert:
        ParserFactoryTestUtils(AliasParserFactory, self).assert_positives([
            'using A = foo', 'using ^ = $$$', 'using A90zaZa = te$t'
        ])

    def test_is_match_returns_false_for_negatives(self):
        # Assert:
        ParserFactoryTestUtils(AliasParserFactory, self).assert_negatives([
            ' using A = foo', 'using A = foo ', 'import A = foo', 'using A = foo bar', 'using A B = foo bar'
        ])


class AliasParserTest(unittest.TestCase):
    def test_can_parse_builtin_as_alias(self):
        for builtin_tuple in BUILTIN_TYPE_TUPLES:
            # Act + Assert:
            SingleLineParserTestUtils(AliasParserFactory, self).assert_parse(
                'using Age = {0}'.format(builtin_tuple[0]),
                ('Age', {'type': 'byte', 'signedness': builtin_tuple[2], 'size': builtin_tuple[1]}))

    def test_alias_names_must_have_type_name_semantics(self):
        # Assert:
        SingleLineParserTestUtils(AliasParserFactory, self).assert_naming(
            'using {0} = uint32',
            VALID_USER_TYPE_NAMES,
            INVALID_USER_TYPE_NAMES)

    def test_cannot_parse_invalid_alias(self):
        # Arrange:
        SingleLineParserTestUtils(AliasParserFactory, self).assert_parse_exceptions([
            'using Hash256 = binary_fixed(2x22)',  # malformed number
            'using Hash256 = binary_fixed(x)',  # malformed number
            'using Age = uint33',  # unknown type
            'using Age = FooBar'  # user type
        ])
